#!/usr/bin/env python3
# -*- coding: utf-8 -*-
""" Wrapper to fetch line database from HITRAN with Astroquery [1]_ 

Notes
-----

Astroquery [R1]_ is itself based on [HAPI]_
    

References
----------

.. [R1] `Astroquery <https://astroquery.readthedocs.io>`_ 


-------------------------------------------------------------------------------

"""

from __future__ import absolute_import
from __future__ import print_function
import radis
from radis.io.hitran import get_molecule, get_molecule_identifier
from radis.misc.cache_files import check_cache_file, get_cache_file, save_to_hdf
from radis.misc import is_float
from astropy import units as u
from astroquery.hitran import Hitran
from os.path import join, isfile, exists
import numpy as np
import pandas as pd
import sys

CACHE_FILE_NAME = 'tempfile_{molecule}_{isotope}_{wmin:.2f}_{wmax:.2f}.h5'

def fetch_astroquery(molecule, isotope, wmin, wmax, verbose=True,
                     cache=True, metadata={}):
    ''' Wrapper to Astroquery [1]_ fetch function to download a line database 

    Notes
    -----

    Astroquery [1]_ is itself based on [HAPI]_

    Parameters
    ----------

    molecule: str, or int
        molecule name or identifier

    isotope: int
        isotope number 

    wmin, wmax: float  (cm-1)
        wavenumber min and max 
    
    Other Parameters
    ----------------

    verbose: boolean
        Default ``True``

    cache: boolean
        if ``True``, tries to find a ``.h5`` cache file in the Astroquery 
        :py:attr:`~astroquery.query.BaseQuery.cache_location`, that would match 
        the requirements. If not found, downloads it and saves the line dataframe 
        as a ``.h5`` file in the Astroquery.

    metadata: dict
        if ``cache=True``, check that the metadata in the cache file correspond
        to these attributes. Arguments ``molecule``, ``isotope``, ``wmin``, ``wmax``
        are already added by default. 

    References
    ----------

    .. [1] `Astroquery <https://astroquery.readthedocs.io>`_ 
    
    See Also
    --------
    
    :py:func:`astroquery.hitran.reader.download_hitran`, 
    :py:func:`astroquery.hitran.reader.read_hitran_file`, 
    :py:attr:`~astroquery.query.BaseQuery.cache_location`

    '''

    # Check input
    if not is_float(molecule):
        mol_id = get_molecule_identifier(molecule)
    else:
        mol_id = molecule
        molecule = get_molecule(mol_id)
    assert is_float(isotope)

    empty_range = False
    
    # If cache, tries to find from Astroquery:
    if cache:    
        # Update metadata with physical properties from the database. 
        metadata.update({'molecule': molecule,
                         'isotope':isotope,
                         'wmin':wmin,
                         'wmax':wmax})
    
        fcache = join(Hitran.cache_location, CACHE_FILE_NAME.format(**{'molecule':molecule,
                                                                       'isotope':isotope,
                                                                       'wmin':wmin,
                                                                       'wmax':wmax}))
        check_cache_file(fcache=fcache, use_cached=cache, metadata=metadata, verbose=verbose)
        if exists(fcache):
            return get_cache_file(fcache, verbose=verbose)

#    tbl = Hitran.query_lines_async(molecule_number=mol_id,
#                                     isotopologue_number=isotope,
#                                     min_frequency=wmin / u.cm,
#                                     max_frequency=wmax / u.cm)
    
#    # Download using the astroquery library
    response = Hitran.query_lines_async(molecule_number=mol_id,
                                         isotopologue_number=isotope,
                                         min_frequency=wmin / u.cm,
                                         max_frequency=wmax / u.cm)
    
    if response.status_code == 404:
        # Maybe there are just no lines for this species in this range
        # In that case we usually end up with errors like:

            # (<class 'Exception'>, Exception('Query failed: 404 Client Error:
            # Not Found for url: http://hitran.org/lbl/api?numax=25000&numin=19000&iso_ids_list=69\n',),
            # <traceback object at 0x7f0967c91708>)

        if response.reason == 'Not Found':
            # Let's bet it's just that there are no lines in this range
            empty_range = True
            if verbose:
                print(('No lines for {0} (id={1}), iso={2} in range {3:.2f}-{4:.2f}cm-1. '.format(
                    molecule, mol_id, isotope, wmin, wmax)))
        else:
            raise ValueError('An error occured during the download of HITRAN files ' +
                             'for {0} (id={1}), iso={2} between {3:.2f}-{4:.2f}cm-1. '.format(
                                 molecule, mol_id, isotope, wmin, wmax) +
                             'Are you online?\n' + 
                             'See details of the error below:\n\n {0}'.format(response.reason))

    # Rename columns from Astroquery to RADIS format
    rename_columns = {'molec_id': 'id',
                      'local_iso_id': 'iso',
                      'nu': 'wav',
                      'sw': 'int',
                      'a': 'A',
                      'gamma_air': 'airbrd',
                      'gamma_self': 'selbrd',
                      'elower': 'El',
                      'n_air': 'Tdpair',
                      'delta_air': 'Pshft',
                      'global_upper_quanta': 'globu',
                      'global_lower_quanta': 'globl',
                      'local_upper_quanta': 'locu',
                      'local_lower_quanta': 'locl',
                      'line_mixing_flag': 'lmix',
                      'gp': 'gp',
                      'gpp': 'gpp',
                      }

    if not empty_range:
#        _fix_astroquery_file_format(filename)
        
        # Note: as of 0.9.16 we're not fixing astroquery_file_format anymore. 
        # maybe we should. 
        
        tbl = Hitran._parse_result(response)
        df = tbl.to_pandas()
        df = df.rename(columns=rename_columns)
    else:
        df = pd.DataFrame(columns=list(rename_columns.values()))

    # Cast type to float64
    cast_type = {'wav': np.float64,
                 'int': np.float64,
                 'A': np.float64,
                 'airbrd': np.float64,
                 'selbrd': np.float64,
                 'El': np.float64,
                 'Tdpair': np.float64,
                 'Pshft': np.float64,
                 }
    for c, typ in cast_type.items():
        df[c] = df[c].astype(typ)

    # cached file mode but cached file doesn't exist yet (else we had returned)
    if cache:
        if verbose:
            print('Generating cached file: {0}'.format(fcache))
        try:
            save_to_hdf(df, fcache, metadata=metadata, version=radis.__version__,
                        key='df', overwrite=True, verbose=verbose)
        except:
            if verbose:
                print(sys.exc_info())
                print('An error occured in cache file generation. Lookup access rights')
            pass

    return df

def _fix_astroquery_file_format(filename):
    '''
    Notes
    -----
    
    On some OS the astroquery lookup function may add extra lines. See:
    https://github.com/astropy/astroquery/issues/1189
    
    In the meantime, we discard all empty lines here. 
    
    '''

    if not isfile(filename):
        raise "{} does not exist "
    with open(filename) as filehandle:
        lines = filehandle.readlines()
        non_empty_lines = [l for l in lines if len(l) > 2]  # > 2 because there may be some line return characters

    if len(lines) != len(non_empty_lines):
        # Re-write file    
        with open(filename, 'w') as filehandle:            
            filehandle.writelines(non_empty_lines)   

if __name__ == '__main__':
    from radis.test.io.test_query import _run_testcases
    _run_testcases(verbose=True)



# -*- coding: utf-8 -*-
"""
Tools to deal with HDF5 cache files
HDF5 cache files are used to cache Energy Database files, and Line Database
files, and yield a much faster access time. 

Routine Listing
---------------


- :func:`~radis.misc.cache_file.check_cache_file`
- :func:`~radis.misc.cache_file.check_not_deprecated`
- :func:`~radis.misc.cache_file.save_to_hdf`

See Also
--------

:py:func:`~radis.io.hitran.hit2df`, 
:py:func:`~radis.io.cdsd.cdsd2df`

-------------------------------------------------------------------------------


"""
# TODO: start using .feather format. Faster than HDF5 for pandas Dataframe, 
# and can be multithreaded.
# see:
# https://gist.github.com/gansanay/4514ec731da1a40d8811a2b3c313f836
# and pd.read_feather(file, nthreads=3)

from __future__ import absolute_import, print_function, division
# Note: don't import unicode_literals because it breaks the df.to_hdf of 
# save_to_hdf because of a stupid unicode/str error in Python 2.7
import os
import h5py
import radis
from warnings import warn
from os.path import exists
from radis import OLDEST_COMPATIBLE_VERSION
from radis.misc.basics import compare_dict, is_float
from radis.misc.printer import printr, printm
import pandas as pd
from packaging.version import parse


class DeprecatedFileError(DeprecationWarning):
    pass

'''str: forces to regenerate cache files that were created in a previous version'''

# Just make sure LAST_BACKWARD_COMPATIBLE_VERSION is valid
assert radis.__version__ >= OLDEST_COMPATIBLE_VERSION

# Utils

def load_h5_cache_file(cachefile, use_cached, metadata, current_version, 
                       last_compatible_version=OLDEST_COMPATIBLE_VERSION,
                       verbose=True):
    ''' Function to load a h5 cache file
    
    Parameters
    ----------
    
    cachefile: str
        cache file path
    
    use_cached: str
        use cache file if value is not ``False``:
        
        - if ``True``, use (and generate if doesnt exist) cache file.
        - if ``'regen'``, delete cache file (if exists) so it is regenerated
        - if ``'force'``, use cache file and raises an error if it doesnt exist
        
        if using the cache file, check if the file is deprecated. If it 
        is deprecated, regenerate the file unless ``'force'`` was used
        (in that case, raise an error)
        
    metadata: dict
        values are compared to cache file attributes. If they dont match,
        the file is considered deprecated. See ``use_cached`` to know
        how to handle deprecated files
        
    current_version: str
        version is compared to cache file version (part of attributes). 
        If current version is superior, a simple warning is triggered.
        
    last_compatible_version: str
        if file version is inferior to this, file is considered deprecated. 
        See ``use_cached`` to know how to handle deprecated files. 
        Default :data:`~radis.OLDEST_COMPATIBLE_VERSION`. 
        
    Returns
    -------
    
    df: pandas DataFrame, or None
        None if no cache file was found, or if it was deleted
        
        
    See Also
    --------
    
    :data:`~radis.OLDEST_COMPATIBLE_VERSION`
        
    '''

    # 1. know if we have to load the file
    if not use_cached:
        return None
    elif use_cached == 'regen' and exists(cachefile):
        os.remove(cachefile)
        if verbose:
            printm('Deleted h5 cache file : {0}'.format(cachefile))
        return None
    
    # 2. check the file is here
    if not exists(cachefile):
        if use_cached == 'force':
            raise ValueError('Cache file {0} doesnt exist'.format(cachefile))
        else:
            return None   # File doesn't exist. It's okay. 
        
    # 3. read file attributes to know if it's deprecated
    try:
        check_not_deprecated(cachefile, metadata, current_version=current_version,
                             last_compatible_version=last_compatible_version)
    # ... if deprecated, raise an error only if 'force'
    except DeprecatedFileError as err:
        if use_cached == 'force':
            raise err
        else:
            if verbose:
                printr('File {0} deprecated:\n{1}\nDeleting it!'.format(
                    cachefile, str(err)))
            os.remove(cachefile)
            return None
    # 4. File is not not deprecated: read the content.
    else:
        df = None
        if verbose>=2:
            printm('Reading cache file ({0})'.format(cachefile))
        try:
            df = pd.read_hdf(cachefile, 'df')
        except KeyError as err:  # An error happened during file reading.
            # Fail safe by deleting cache file (unless we explicitely wanted it
            # with 'force')
            if use_cached == 'force':
                raise
            else:
                if verbose:
                    printr('An error happened during cache file reading ' +
                           '{0}:\n{1}\n'.format(cachefile, str(err)) +
                           'Deleting cache file to regenerate it')
                os.remove(cachefile)
                df = None

    return df

def get_cache_file(fcache, verbose=True):
    ''' Load HDF5 cache file 
    
    Parameters
    ----------
    
    fcache: str
        file name
    
    Other Parameters
    ----------------
    
    verbose: bool
        If >=2, also warns if non numeric values are present (it would make 
        calculations slower)

    Notes
    -----
    
    we could start using FEATHER format instead. See notes in cache_files.py 
    '''

    # Load file
    
    df = pd.read_hdf(fcache, 'df')
    
    # Check file
    
    # ... 'object' columns slow everything down (not fixed format strings!)
    if verbose >= 2:
        _warn_if_object_columns(df, fcache)
    
    return df

def check_cache_file(fcache, use_cached=True, metadata={}, verbose=True):
    ''' Quick function that check status of cache file generated by RADIS:

    Parameters
    ----------
    
    fcache: str
        cache file name
        
    use_cached: ``True``, ``False``, ``'force'``, ``'regen'``
        see notes below. Default ``True``.
        
    metadata: dict
        attributes to check
        
    verbose: boolean
        print stuff
    
    Notes
    -----
    
    The function first checks the existence of ``fcache``. What is does depends
    on the value of ``use_cached``:

    - if ``True``, check it exists. 
    - if ``'regen'``, delete cache file to regenerate it later. 
    - if ``'force'``, raise an error if file doesnt exist. 

    Then look if it is deprecated (we just look at the attributes, the file 
    is never fully read). Deprecation is done by :py:func:`~radis.misc.cache_files.check_not_deprecated`
    comparing the ``metadata=`` content.

    - if deprecated, deletes it to regenerate later unless 'force' was used
    
    See Also
    --------
    
    :py:func:`~radis.misc.cache_files.check_not_deprecated`
    '''

    # Test existence of file:
    if not use_cached:
        return                   # we dont want a cache file, no need to test it
    elif use_cached == 'regen':
        if exists(fcache):
            os.remove(fcache)
            if verbose:
                print(('Deleted h5 cache file : {0}'.format(fcache)))
    elif use_cached == 'force':
        if not exists(fcache):
            raise ValueError('Cache file {0} doesnt exist'.format(fcache))
    else:  # use_cached == True
        pass    # just use the file as is

    # If file is still here, test if it is deprecated:
    # (we just read the attributes, the file is never fully read)
    if exists(fcache):
        if verbose:
            print(('Using cache file: {0}'.format(fcache)))
        try:
            check_not_deprecated(fcache, metadata=metadata, current_version=radis.__version__,
                                 last_compatible_version=OLDEST_COMPATIBLE_VERSION)
        except DeprecatedFileError as err:
            if use_cached == 'force':
                raise
            else:   # delete file to regenerate it in the end of the script
                if verbose:
                    printr('File {0} deprecated:\n{1}\nDeleting it!'.format(
                        fcache, str(err)))
                os.remove(fcache)

    return


def check_not_deprecated(file, metadata, current_version=None, 
                         last_compatible_version=OLDEST_COMPATIBLE_VERSION):
    ''' Make sure cache file is not deprecated: checks that ``metadata`` is the same,
    and that the version under which the file was generated is valid.

    Parameters
    ----------

    file: str
        a `` .h5`` cache file for Energy Levels 

    metadata: dict
        list of variables used to create the file. If the values dont match,
        an error is raised. 

    current_version: str, or ``None``
        current version number. If the file was generated in a previous version
        a warning is raised. If ``None``, current version is read from 
        :data:`radis.__version__`.

    last_backward_compatible_version: str
        If the file was generated in a non-compatible version, an error is raised.
        Default :py:data:`~radis.OLDEST_COMPATIBLE_VERSION`

    '''

    # Get attributes (metadata+version)
    hf = h5py.File(file, 'r')
    try:
        attrs = dict(hf.attrs)
    except OSError:
        attrs = {}
    finally:
        hf.close()

    # Raise an error if version is not found
    try:
        file_version = attrs.pop('version')
    except KeyError:
        raise DeprecatedFileError('File {0} has been generated in a deprecated '.format(file) +
                                  'version. Delete it to regenerate it on next run')

    # Get current version
    if current_version is None:
        current_version = radis.__version__

    # If file version is anterior to a major change
    # ... Update here versions afterwhich Deprecated cache file is not safe
    # ... (example: a key name was changed)
    if parse(file_version) < parse(last_compatible_version):
        raise DeprecatedFileError('File {0} has been generated in a deprecated '.format(file) +
                                  'version ({0}). Oldest compatible version is {1}. '.format(
            file_version, last_compatible_version) +
            'Delete the file to regenerate it on next run')

    # If file version is outdated: Warning, but no error
    if parse(current_version) > parse(file_version):
        warn(DeprecationWarning('File {0} has been generated in '.format(file) +
                                'a deprecated version ({0}) compared to current ({1})'.format(
            file_version, current_version) +
            '. Delete it to regenerate it on next run'))
        out = False
    elif parse(current_version) == parse(file_version):
        out = True
    else:
        raise ValueError('Cache file ({0}) generated with a future version ({1} > {2})? '.format(
                            file, file_version, current_version)+\
                         "Do you own a DeLorean? Delete the file manually if you understand what happened")

    # Compare metadata
    metadata = _h5_compatible(metadata)
    out, compare_string = compare_dict(
        metadata, attrs, verbose=False, return_string=True)
    if out != 1:
        raise DeprecatedFileError('Metadata in file {0} dont match '.format(file) +
                                  'expected values. See comparison below:' +
                                  '\n\tExpected\tFile\n{0}'.format(compare_string))

    return out


def _h5_compatible(a_dict):
    ''' Make dictionary ``a_dict`` h5 compatible '''
    out = {}
    for k, v in a_dict.items():
        if v is None:
            continue   # dont store None
        elif is_float(v):
            out[k] = v
        else:
            out[k] = str(v)   # convert to str
    return out

def _warn_if_object_columns(df, fname):
    '''  'object' columns slow everything down (not fixed format strings!) '''
    objects = [k for k,v in df.dtypes.items() if v == object]
    if len(objects) > 0:
        warn('Dataframe in {0} contains `object` format columns: {1}. '.format(
                fname, objects)+'Operations will be slower. Try to convert them to numeric.')


def save_to_hdf(df, fname, metadata, version=None, key='df', overwrite=True,
                verbose=True):
    ''' Save energy levels to HDF5 file. Add metadata and version 

    Parameters
    ----------

    df: a pandas DataFrame
        data will be stored in the key ``'df'`` 

    fname: str
        ``.h5`` file where to store. 

    metadata: dict
         dictionary of values that were used to generate the DataFrame. Metadata
         will be asked again on file load to ensure it hasnt changed. ``None`` 
         values will not be stored.

    version: str, or ``None``
        file version. If ``None``, the current :data:`radis.__version__` is used. 
        On file loading, a warning will be raised if the current version is 
        posterior, or an error if the file version is set to be uncompatible.
        See :py:data:`~radis.OLDEST_COMPATIBLE_VERSION`

    key: str
        dataset name. Default ``'df'`` 

    overwrite: boolean
        if ``True``, overwrites file. Else, raise an error if it exists.

    Other Parameters
    ----------------
    
    verbose: bool
        If >=2, also warns if non numeric values are present (it would make 
        calculations slower)

    Notes
    -----

    ``None`` values are not stored

    '''

    # Check file
    assert fname.endswith('.h5')
    assert 'version' not in metadata
    # ... 'object' columns slow everything down (not fixed format strings!)
    if verbose >= 2:
        _warn_if_object_columns(df, fname)

    # Update metadata format
    metadata = _h5_compatible(metadata)

    # Overwrite file
    if exists(fname) and not overwrite:
        raise ValueError('File exist: {0}'.format(fname))
    hf = h5py.File(fname, 'w')

    try:

        # Start by adding version
        if version is None:
            version = radis.__version__
        hf.attrs['version'] = version

        # Add metadata
        for k, v in metadata.items():
            hf.attrs[k] = v

    except:
        raise

    finally:
        hf.close()

    # now export dataframe
    df.to_hdf(fname, key, format='fixed', mode='a',
              complevel=1, complib='blosc')
#    df.to_hdf(fname, 'df', format='fixed', mode='a')  # would be 10-20% faster, but take 2x more space


def filter_metadata(arguments, discard_variables=['self', 'verbose']):
    ''' Filter arguments (created with  ``locals()`` at the beginning
    of the script) to extract metadata. Metadata is stored as attributes
    in the cached file:

    - remove variables in ``discard_variables``
    - remove variables that start with ``'_'``
    - remove varibles whose value is ``None``

    Parameters
    ----------

    arguments: dict
        list of local variables. For instance::

            arguments = locals()

    discard_variables: list of str
        variable names to discard

    Returns
    -------

    metadata: dict
        a (new) dictionary built from arguments by removing ``discard_variables``
        and variables starting with ``'_'``

    Examples
    --------

    How to get only function argument::

        def some_function(*args):
            metadata = locals()     # stores only function arguments because it's the first line

            ...

            metadata = filter_metadata(metadata)
            save_to_hdf(df, fname, metadata=metadata)

            ...

    '''

    metadata = {k: v for (k, v) in arguments.items() if not k.startswith('_')}
    metadata = {k: v for (k, v) in metadata.items()
                if k not in discard_variables}
    metadata = {k: v for (k, v) in metadata.items() if v is not None}

    return metadata



# -*- coding: utf-8 -*-
"""
Determine Gas composition using [CANTERA]_

References
----------

.. [CANTERA]  D. G. Goodwin, H. K. Moffat, R. L. Speth, Cantera: An Object-oriented Software
              Toolkit for Chemical Kinetics, Thermodynamics, and Transport Processes,
              http://www.cantera.org, doi:10.5281/zenodo.170284, 2017.

-------------------------------------------------------------------------------

"""

from __future__ import absolute_import
from radis.misc.utils import NotInstalled
try:
    import cantera as ct
except:
    ct = NotInstalled('cantera', 'Cantera is needed to calculate equilibrium mole fractions'+\
                      '. Install with  `pip install cantera`')
    
def get_eq_mole_fraction(initial_mixture, T_K, p_Pa):
    ''' Returns mole fraction at temperature T, using the 
    [CANTERA]_ :py:meth:`~cantera.ThermoPhase.equilibrate` function. 
    
    Parameters
    ----------
    
    initial_mixture: str
        Gas composition. Example::
            
             'N2:0.79, O2:0.21, CO2:363e-6'
             
        Or::
            
             'CO2:1'
             
    T_K: float (K)
        temperature (Kelvin) to calculate equilibrium
        
    P_Pa: float (Pa)
        temperature (Pascal) to calculate equilibrium
        
    Examples
    --------
    
    Calculate equilibrium mixture of CO2 at 2000 K, 1 atm::
        
        get_eq_mole_fraction('CO2:1', 2000, 101325)
        
        >>> {'C': 1.7833953335281855e-19,
            'CO': 0.01495998583472384,
            'CO2': 0.9775311634424326,
            'O': 5.7715610124613225e-05,
            'O2': 0.007451135112719029}
    
    References
    ----------
    
    [CANTERA]_
    
    '''

    # %% Init Cantera
    g = ct.Solution('gri30.xml')
    g.TPX = T_K, p_Pa, initial_mixture
    
    # Minimize Gibbs:
    g.equilibrate('TP')
    
    # Returns
    return g.mole_fraction_dict()
    



# -*- coding: utf-8 -*-
"""
Spectrum-class module 
"""

from .spectrum import (Spectrum, is_spectrum)
from .models import (calculated_spectrum, experimental_spectrum,
                       transmittance_spectrum)
from .compare import (plot_diff, get_diff, get_distance, get_ratio, get_residual,
                      get_residual_integral)
from .operations import (Transmittance, Radiance, Radiance_noslit, Transmittance_noslit,
                         PerfectAbsorber, get_baseline, sub_baseline)



# -*- coding: utf-8 -*-
"""
Created on Sun Aug  5 14:26:44 2018

@author: erwan
"""

from __future__ import print_function, absolute_import, division, unicode_literals
from radis.tools.database import load_spec
from radis.spectrum.operations import crop, Radiance_noslit, add_constant, multiply
from radis.test.utils import getTestFile
import pytest

@pytest.mark.fast
def test_crop(verbose=True, *args, **kwargs):
    ''' Test that update can correctly recompute missing quantities '''

    # 1) A crop example in the same unit as the one stored

    # Work with a Spectrum object that was generated by Specair
    s = load_spec(getTestFile('N2C_specair_380nm.spec'))
    # Focus on N2(C->B), v'=0, v''=2 band:
    s.crop(376, 381, 'nm')
    
    w = s.get_wavelength()
    assert w.min() >= 360
    assert w.max() <= 382
    
    # 2) A crop example in different unit as the one stored

    # Work with a Spectrum object that was generated by Specair
    s1 = load_spec(getTestFile('CO_Tgas1500K_mole_fraction0.01.spec'), binary=True)
    # Focus on N2(C->B), v'=0, v''=2 band:
    s1.crop(4530, 4533, 'nm')
    
    w = s1.get_wavelength()
    assert w.min() >= 4530
    assert w.max() <= 4533
    
    return True


@pytest.mark.fast
def test_cut_recombine(verbose=True, *args, **kwargs):
    ''' 
    Use :func:`~radis.spectrum.operations.crop` and :func:`~radis.los.slabs.MergeSlabs`
    to cut a Spectrum and recombine it
    
    Assert we still get the same spectrum at the end
    '''
    from radis import MergeSlabs

    s = load_spec(getTestFile('CO_Tgas1500K_mole_fraction0.01.spec'), binary=True)
    # Cut in half
    cut = 2177
    s1 = crop(s, 2000, cut, 'cm-1', inplace=False)
    s2 = crop(s, cut, 2300, 'cm-1', inplace=False)

    # Recombine
    s_new = MergeSlabs(s1, s2, resample='full', out='transparent')
    
    # Compare
    assert s.compare_with(s_new, spectra_only=True, plot=False, verbose=verbose)
    
@pytest.mark.fast
def test_invariants(*args, **kwargs):
    ''' Ensures adding 0 or multiplying by 1 does not change the spectra '''
    from radis import load_spec
    from radis.test.utils import getTestFile
    
    s = load_spec(getTestFile("CO_Tgas1500K_mole_fraction0.01.spec"))
    s.update()
    s = Radiance_noslit(s)

    assert s.compare_with(add_constant(s, 0, 'W/cm2/sr/nm'), 
                          plot=False, spectra_only='radiance_noslit')
    assert s.compare_with(multiply(s, 1), 
                          plot=False, spectra_only='radiance_noslit')
    
    assert 3*s/3 == s
    assert (1+s)-1 == s
    

@pytest.mark.fast
def test_operations_inplace(verbose=True, *args, **kwargs):
    
    from radis.spectrum.operations import Radiance_noslit

    s = load_spec(getTestFile('CO_Tgas1500K_mole_fraction0.01.spec'), binary=True)
    s.update('radiance_noslit', verbose=False)
    s = Radiance_noslit(s)
    
    # Add 1, make sure it worked
    I_max = s.get('radiance_noslit')[1].max()    
    s += 1
    assert s.get('radiance_noslit')[1].max() == I_max + 1
    if verbose:
        print('test_operations: s += 1: OK')
    
    # Multiply, make sure it worked
    I_max = s.get('radiance_noslit')[1].max()    
    s *= 10
    assert s.get('radiance_noslit')[1].max() == 10*I_max
    if verbose:
        print('test_operations: s *= 10: OK')
    
def _run_testcases(verbose=True, *args, **kwargs):
    ''' Test procedures
    '''

    test_crop(verbose=verbose, *args, **kwargs)
    test_cut_recombine(verbose=verbose, *args, **kwargs)
    test_invariants(verbose=verbose, *args, **kwargs)
    test_operations_inplace(verbose=verbose, *args, **kwargs)
    
    return True

if __name__ == '__main__':
    
    _run_testcases()
    